"""Input event models for the Python SDK."""

from __future__ import annotations

from typing import Annotated, Any, Literal, TypeAlias

from pydantic import Field

from ._base import TerseModel


class InputEvent(TerseModel):
    integration_type: str
    event_type: str = "unknown"
    formatted_content: str = ""
    debug_log: str = ""
    metadata: dict[str, Any] | None = None


class GithubRepository(TerseModel):
    id: int
    name: str
    owner: str
    default_branch: str = "main"


class GithubUser(TerseModel):
    login: str
    email: str | None = None


class GithubFileDiff(TerseModel):
    filename: str
    diff: str


class GithubCommit(TerseModel):
    sha: str
    message: str
    file_diffs: list[GithubFileDiff] = Field(default_factory=list)


class GithubPRRef(TerseModel):
    ref: str
    sha: str


class GithubPRData(TerseModel):
    number: int
    title: str
    body: str | None = None
    state: Literal["open", "closed"]
    merged: bool = False
    head: GithubPRRef
    base: GithubPRRef
    author: GithubUser
    url: str


class GithubEventMetadata(TerseModel):
    repository: GithubRepository | None = None
    sender: GithubUser | None = None
    commits: list[GithubCommit] = Field(default_factory=list)
    pull_request: GithubPRData | None = None
    branch: str | None = None


class GithubInputEvent(InputEvent):
    integration_type: Literal["github"] = "github"
    metadata: GithubEventMetadata | None = None


class SlackInputEvent(InputEvent):
    integration_type: Literal["slack"] = "slack"


class LinearInputEvent(InputEvent):
    integration_type: Literal["linear"] = "linear"


class AtlassianInputEvent(InputEvent):
    integration_type: Literal["atlassian"] = "atlassian"


class GmailInputEvent(InputEvent):
    integration_type: Literal["gmail"] = "gmail"


class NotionInputEvent(InputEvent):
    integration_type: Literal["notion"] = "notion"


class FigmaInputEvent(InputEvent):
    integration_type: Literal["figma"] = "figma"


class PosthogInputEvent(InputEvent):
    integration_type: Literal["posthog"] = "posthog"


class DatadogInputEvent(InputEvent):
    integration_type: Literal["datadog"] = "datadog"


class TerseInputEvent(InputEvent):
    integration_type: Literal["terse"] = "terse"


class CronJobInputEvent(InputEvent):
    integration_type: Literal["cron_job"] = "cron_job"


class LaunchDarklyInputEvent(InputEvent):
    integration_type: Literal["launchdarkly"] = "launchdarkly"


class WorkOSEventUser(TerseModel):
    id: str
    email: str
    first_name: str | None = None
    last_name: str | None = None
    email_verified: bool
    profile_picture_url: str | None = None


class WorkOSEventMembership(TerseModel):
    id: str
    user_id: str
    organization_id: str
    role: dict[str, str]
    status: str


class WorkOSEventInvitation(TerseModel):
    id: str
    email: str
    organization_id: str
    inviter_email: str | None = None
    state: str
    accepted_at: str | None = None


class WorkOSEventMetadata(TerseModel):
    event_id: str | None = None
    created_at: str | None = None
    user: WorkOSEventUser | None = None
    membership: WorkOSEventMembership | None = None
    invitation: WorkOSEventInvitation | None = None


class WorkOSInputEvent(InputEvent):
    integration_type: Literal["workos"] = "workos"
    metadata: WorkOSEventMetadata | None = None


class AttioInputEvent(InputEvent):
    integration_type: Literal["attio"] = "attio"


class SnowflakeInputEvent(InputEvent):
    integration_type: Literal["snowflake"] = "snowflake"


class SerializedEventInputEvent(InputEvent):
    """Fallback event for integrations without a dedicated typed model yet."""


KnownInputEvent: TypeAlias = Annotated[
    GithubInputEvent
    | SlackInputEvent
    | LinearInputEvent
    | AtlassianInputEvent
    | GmailInputEvent
    | NotionInputEvent
    | FigmaInputEvent
    | PosthogInputEvent
    | DatadogInputEvent
    | TerseInputEvent
    | CronJobInputEvent
    | LaunchDarklyInputEvent
    | WorkOSInputEvent
    | AttioInputEvent
    | SnowflakeInputEvent,
    Field(discriminator="integration_type"),
]

AnyInputEvent: TypeAlias = KnownInputEvent | SerializedEventInputEvent


__all__ = [
    "AnyInputEvent",
    "AtlassianInputEvent",
    "AttioInputEvent",
    "CronJobInputEvent",
    "DatadogInputEvent",
    "FigmaInputEvent",
    "GithubCommit",
    "GithubEventMetadata",
    "GithubFileDiff",
    "GithubInputEvent",
    "GithubPRData",
    "GithubPRRef",
    "GithubRepository",
    "GithubUser",
    "GmailInputEvent",
    "InputEvent",
    "KnownInputEvent",
    "LaunchDarklyInputEvent",
    "LinearInputEvent",
    "NotionInputEvent",
    "PosthogInputEvent",
    "SerializedEventInputEvent",
    "SlackInputEvent",
    "SnowflakeInputEvent",
    "TerseInputEvent",
    "WorkOSEventInvitation",
    "WorkOSEventMembership",
    "WorkOSEventMetadata",
    "WorkOSEventUser",
    "WorkOSInputEvent",
]



"""Run history and approval models."""

from __future__ import annotations

from ._base import _CamelModel
from .enums import (
    ApprovalActionType,
    ApprovalRequestStatus,
    ChangeEventType,
    ConfigType,
    EntityType,
    IntegrationType,
    RunHistoryActionType,
    RunHistoryDecisionAction,
    RunHistoryStatus,
)


class OutputItem(_CamelModel):
    output_item_id: str
    output_item_type: ConfigType


class RunHistoryTrigger(_CamelModel):
    event: str
    integration: IntegrationType
    source: str
    subheader: str | None = None
    title: str | None = None
    url: str | None = None


class ApprovalAction(_CamelModel):
    deepLink: str
    label: str
    type: ApprovalActionType


class ApprovalRequest(_CamelModel):
    actions: list[ApprovalAction]
    agentId: str
    icon: IntegrationType
    id: str
    runId: str
    status: ApprovalRequestStatus
    subheader: str
    timestamp: str
    title: str


class ChangedItem(_CamelModel):
    change_event_type: ChangeEventType
    id: str
    type_name: EntityType


class RunHistoryAction(_CamelModel):
    action: str
    details: str
    integration: IntegrationType
    isReadOnly: bool | None = None
    output_items: list[OutputItem] | None = None
    step_id: str | None = None
    target: str
    type: RunHistoryActionType
    url: str | None = None


class RunHistoryDecision(_CamelModel):
    action: RunHistoryDecisionAction
    reasoning: str


class RunHistoryRecord(_CamelModel):
    actions: list[RunHistoryAction] | None = None
    agentId: str
    decision: RunHistoryDecision
    filtered: bool
    id: str
    isManuallyTriggered: bool
    status: RunHistoryStatus
    timestamp: str
    trigger: RunHistoryTrigger


__all__ = [
    "ApprovalAction",
    "ApprovalRequest",
    "ChangedItem",
    "OutputItem",
    "RunHistoryAction",
    "RunHistoryDecision",
    "RunHistoryRecord",
    "RunHistoryTrigger",
]



# ruff: noqa: E501
from __future__ import annotations

import json
import os
import sys
from types import ModuleType, SimpleNamespace
from typing import Any
from unittest.mock import patch

import httpx
import pytest
import terse_sdk.runtime as runtime_module
from terse_sdk import (
    CronJobInputEvent,
    EventType,
    MissingApiKeyError,
    SdkAgentStreamEventDone,
    SdkAgentStreamEventFinalOutput,
    SdkAgentStreamEventText,
    SdkAgentStreamEventToolCallCompleted,
    SerializedEventInputEvent,
    Terse,
    TerseAgent,
    TerseApiError,
    clear_job_registry,
    deserialize_input_event,
    execute_registered_job,
    get_job_registry,
)


@pytest.fixture(autouse=True)
def clear_registry() -> None:
    clear_job_registry()
    yield
    clear_job_registry()


class _FakeSSEEvent:
    def __init__(self, data: str) -> None:
        self.data = data


class _FakeEventSource:
    def __init__(self, events: list[str], *, path: str = "/sdk/agent-run") -> None:
        self.response = _json_response(
            200,
            {"ok": True},
            headers={"Content-Type": "text/event-stream"},
            path=path,
        )
        self._events = events

    def __enter__(self) -> _FakeEventSource:
        return self

    def __exit__(self, exc_type: object, exc: object, tb: object) -> bool:
        return False

    def iter_sse(self):
        for event in self._events:
            yield _FakeSSEEvent(event)


class _FakeClient:
    def __init__(self, response: httpx.Response | None = None) -> None:
        self.response = response
        self.calls: list[dict[str, Any]] = []

    def __enter__(self) -> _FakeClient:
        return self

    def __exit__(self, exc_type: object, exc: object, tb: object) -> bool:
        return False

    def post(self, url: str, *, headers: dict[str, str], json: dict[str, object]) -> httpx.Response:
        self.calls.append({"url": url, "headers": headers, "json": json})
        if self.response is None:
            raise AssertionError("No response configured")
        return self.response


def test_job_registration_and_registry_clear() -> None:
    app = Terse()

    @app.job(name="demo-job")
    def demo(event: CronJobInputEvent, agent: TerseAgent) -> None:
        _ = (event, agent)

    registry = get_job_registry()
    assert "demo-job" in registry
    assert registry["demo-job"].handler is demo

    clear_job_registry()
    assert get_job_registry() == {}


def test_execute_registered_job_supports_sync_and_async_callables() -> None:
    sync_calls: list[str] = []
    async_calls: list[str] = []
    app = Terse()

    @app.job(name="sync-job")
    def sync_handler(event: CronJobInputEvent, agent: TerseAgent) -> None:
        _ = agent
        sync_calls.append(event.formatted_content)

    async def allow_async(event: CronJobInputEvent) -> bool:
        return event.event_type == "manual"

    @app.job(name="async-job", filter=allow_async)
    async def async_handler(event: CronJobInputEvent, agent: TerseAgent) -> None:
        _ = agent
        async_calls.append(event.debug_log)

    event = CronJobInputEvent(
        event_type="manual",
        formatted_content="hello",
        debug_log="world",
    )

    sync_job = get_job_registry()["sync-job"]
    async_job = get_job_registry()["async-job"]

    assert not execute_registered_job(sync_job, event, agent=TerseAgent())
    assert sync_calls == ["hello"]

    assert not execute_registered_job(async_job, event, agent=TerseAgent())
    assert async_calls == ["world"]


def test_execute_registered_job_returns_true_when_filter_skips() -> None:
    calls: list[str] = []
    app = Terse()

    def never(event: CronJobInputEvent) -> bool:
        _ = event
        return False

    @app.job(name="demo-job", filter=never)
    def demo(event: CronJobInputEvent, agent: TerseAgent) -> None:
        _ = (event, agent)
        calls.append("ran")

    skipped = execute_registered_job(
        get_job_registry()["demo-job"],
        CronJobInputEvent(event_type="manual", formatted_content="demo", debug_log="demo"),
        agent=TerseAgent(),
    )

    assert skipped
    assert calls == []


def test_deserialize_input_event_supports_camel_case_payloads() -> None:
    event = deserialize_input_event(
        {
            "integrationType": "cron_job",
            "eventType": "manual",
            "formattedContent": "Scheduled job",
            "debugLog": "cron",
        }
    )

    assert isinstance(event, CronJobInputEvent)
    assert event.integration_type == "cron_job"
    assert event.formatted_content == "Scheduled job"


def test_deserialize_input_event_falls_back_for_unknown_integrations() -> None:
    event = deserialize_input_event(
        {
            "integrationType": "unknown_service",
            "eventType": "manual",
            "formattedContent": "Unknown",
            "debugLog": "unknown",
        }
    )

    assert isinstance(event, SerializedEventInputEvent)
    assert event.integration_type == "unknown_service"


def test_agent_execute_tool_includes_session_and_run_headers() -> None:
    fake_client = _FakeClient(
        _json_response(
            200,
            {"success": True, "result": {"ok": True}},
            path="/sdk/tool-execute",
        )
    )

    with (
        patch.dict(os.environ, {"TERSE_API_KEY": "terse_test_key", "TERSE_RUN_ID": "run_123"}, clear=False),
        patch("terse_sdk.runtime.httpx.Client", return_value=fake_client),
    ):
        result = TerseAgent(session_id="session_123").execute_tool("demo_tool", {"value": 1})

    assert result == {"ok": True}
    assert fake_client.calls[0]["headers"]["X-Terse-Session-Id"] == "session_123"
    assert fake_client.calls[0]["headers"]["X-Terse-Run-Id"] == "run_123"
    assert fake_client.calls[0]["json"]["toolName"] == "demo_tool"


def test_agent_execute_tool_requires_api_key() -> None:
    with patch.dict(os.environ, {"TERSE_API_KEY": ""}, clear=False), pytest.raises(MissingApiKeyError):
        TerseAgent().execute_tool("demo_tool")


def test_agent_tools_lazy_attach_from_generated_module() -> None:
    created_agents: list[TerseAgent] = []
    fake_tools = SimpleNamespace(snowflake="snowflake-tools")
    generated_module = ModuleType("terse_generated")

    def create_tools(agent: TerseAgent) -> object:
        created_agents.append(agent)
        return fake_tools

    generated_module.create_tools = create_tools  # type: ignore[attr-defined]

    with patch.dict(sys.modules, {"terse_generated": generated_module}, clear=False):
        agent = TerseAgent()
        assert agent.tools is fake_tools
        assert agent.tools is fake_tools

    assert created_agents == [agent]


def test_agent_tools_raise_clear_error_when_generated_module_is_missing() -> None:
    with (
        patch("terse_sdk.runtime._resolve_generated_tools_factory", return_value=None),
        pytest.raises(AttributeError),
    ):
        _ = TerseAgent().tools


def test_agent_run_streams_events_and_serializes_event_payload() -> None:
    captured: dict[str, object] = {}
    stream = _FakeEventSource(
        [
            SdkAgentStreamEventText(type="text", text="hello").model_dump_json(),
            SdkAgentStreamEventToolCallCompleted(
                type="tool_call_completed",
                toolCallCompleted=json.dumps({"tool": "demo_tool", "status": "completed"}),
            ).model_dump_json(),
            SdkAgentStreamEventFinalOutput(type="final_output", finalOutput="done").model_dump_json(),
            SdkAgentStreamEventDone(type="done").model_dump_json(),
        ]
    )

    def fake_connect_sse(
        client: object,
        method: str,
        url: str,
        *,
        headers: dict[str, str],
        json: dict[str, object],
    ) -> _FakeEventSource:
        captured.update({"method": method, "url": url, "headers": headers, "json": json})
        return stream

    with (
        patch.dict(os.environ, {"TERSE_API_KEY": "terse_test_key"}, clear=False),
        patch("terse_sdk.runtime.connect_sse", side_effect=fake_connect_sse),
    ):
        events = list(
            TerseAgent().run(
                "hello",
                CronJobInputEvent(
                    event_type="manual",
                    formatted_content="Scheduled event",
                    debug_log="cron",
                ),
            )
        )

    assert len(events) == 3
    assert events[-1].type == EventType.FINAL_OUTPUT
    assert events[-1].finalOutput == "done"
    assert captured["method"] == "POST"
    assert captured["headers"]["Authorization"] == "Bearer terse_test_key"
    assert captured["json"]["event"]["integrationType"] == "cron_job"
    assert captured["json"]["event"]["formattedContent"] == "Scheduled event"


def test_agent_run_raises_on_failed_tool_call() -> None:
    stream = _FakeEventSource(
        [
            SdkAgentStreamEventToolCallCompleted(
                type="tool_call_completed",
                toolCallCompleted=json.dumps({"tool": "demo_tool", "status": "failed"}),
            ).model_dump_json(),
            SdkAgentStreamEventDone(type="done").model_dump_json(),
        ]
    )

    with (
        patch.dict(os.environ, {"TERSE_API_KEY": "terse_test_key"}, clear=False),
        patch("terse_sdk.runtime.connect_sse", return_value=stream),
        pytest.raises(TerseApiError),
    ):
        list(TerseAgent().run("hello"))


def test_agent_run_and_wait_returns_final_output() -> None:
    with patch.object(
        TerseAgent,
        "run",
        return_value=iter(
            [
                SdkAgentStreamEventText(type="text", text="thinking"),
                SdkAgentStreamEventFinalOutput(type="final_output", finalOutput="done"),
            ]
        ),
    ):
        result = TerseAgent().run_and_wait("hello")

    assert result == "done"


def test_agent_run_and_wait_returns_none_when_no_final_output_arrives() -> None:
    with patch.object(TerseAgent, "run", return_value=iter([])):
        result = TerseAgent().run_and_wait("hello")

    assert result is None


def test_agent_run_and_wait_propagates_errors() -> None:
    with patch.object(TerseAgent, "run", side_effect=TerseApiError("boom")), pytest.raises(TerseApiError):
        TerseAgent().run_and_wait("hello")


def test_assert_sse_response_reads_streaming_json_error_payload() -> None:
    response = _streaming_json_response(
        200,
        {"success": False, "error": "backend unavailable"},
        path="/sdk/agent-run",
    )

    with pytest.raises(TerseApiError) as exc_info:
        runtime_module._assert_sse_response(response, "/sdk/agent-run")

    assert "backend unavailable" in str(exc_info.value)


def test_assert_sse_response_reads_streaming_error_detail_on_http_error() -> None:
    response = _streaming_json_response(
        401,
        {"error": "unauthorized"},
        path="/sdk/agent-run",
    )

    with pytest.raises(TerseApiError) as exc_info:
        runtime_module._assert_sse_response(response, "/sdk/agent-run")

    assert "unauthorized" in str(exc_info.value)


def _json_response(
    status_code: int,
    payload: object,
    *,
    headers: dict[str, str] | None = None,
    path: str,
) -> httpx.Response:
    return httpx.Response(
        status_code,
        headers={"Content-Type": "application/json", **(headers or {})},
        content=json.dumps(payload).encode("utf-8"),
        request=httpx.Request("POST", f"https://example.com{path}"),
    )


def _streaming_json_response(
    status_code: int,
    payload: object,
    *,
    headers: dict[str, str] | None = None,
    path: str,
) -> httpx.Response:
    return httpx.Response(
        status_code,
        headers={"Content-Type": "application/json", **(headers or {})},
        stream=httpx.ByteStream(json.dumps(payload).encode("utf-8")),
        request=httpx.Request("POST", f"https://example.com{path}"),
    )



"""Attio-related models."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel

from ._base import _CamelModel


class AttioAttribute(BaseModel):
    api_slug: str | None = None
    is_required: bool | None = None
    is_unique: bool | None = None
    title: str | None = None
    type: str | None = None


class AttioObjectWithAttributes(_CamelModel):
    api_slug: str
    attributes: list[AttioAttribute] | None = None
    plural_noun: str
    singular_noun: str


class AttioRecordIdentifier(BaseModel):
    object_id: str | None = None
    record_id: str | None = None
    workspace_id: str | None = None


class AttioRecord(BaseModel):
    created_at: str | None = None
    id: AttioRecordIdentifier | None = None
    values: dict[str, Any] | None = None
    web_url: str | None = None


class Repository(_CamelModel):
    id: float
    name: str
    owner: str


__all__ = [
    "AttioAttribute",
    "AttioObjectWithAttributes",
    "AttioRecord",
    "AttioRecordIdentifier",
    "Repository",
]



"""SDK request and response DTOs."""

from __future__ import annotations

from typing import Any, Literal

from ._base import _CamelModel
from .enums import ConfigType, IntegrationType


class ApiToken(_CamelModel):
    createdAt: str
    id: str
    lastUsedAt: str | None
    name: str
    tokenPrefix: str


class ApiTokenCreateResponse(_CamelModel):
    rawToken: str
    token: ApiToken


class PartialSdkAgentRunEventPayload(_CamelModel):
    debugLog: str | None = None
    formattedContent: str | None = None
    integrationType: IntegrationType | None = None


class SdkAgentRunEventPayload(_CamelModel):
    debugLog: str
    formattedContent: str
    integrationType: IntegrationType


class SdkAgentRunOptionsPayload(_CamelModel):
    maxTurns: float | None = None
    requireApproval: bool | None = None


class Contract(_CamelModel):
    responseMode: Literal["streaming"]
    supportsInterruptions: bool


class Options(_CamelModel):
    maxTurns: float
    requireApproval: bool


class RemovedItem(_CamelModel):
    id: str
    name: str


class Result(_CamelModel):
    automationId: str
    isUpdate: bool
    jobName: str


class SdkAgentSkillPayload(_CamelModel):
    config: dict[str, Any]
    configType: ConfigType


class TriggerPayload(_CamelModel):
    config: dict[str, Any]
    integrationId: str
    integrationType: IntegrationType


class SdkDeployTrigger(_CamelModel):
    config: dict[str, Any]
    configType: str
    integrationId: str
    integrationType: str


class SdkDeployJob(_CamelModel):
    jobName: str
    triggers: list[SdkDeployTrigger]
    webhookURL: str | None = None


class SdkDeployRequestBody(_CamelModel):
    jobs: list[SdkDeployJob]
    sourceZipBase64: str


class SdkDeployResponseBody(_CamelModel):
    details: str | None = None
    error: str | None = None
    removed: list[RemovedItem]
    results: list[Result]
    success: bool


class SerializedEvent(_CamelModel):
    debugLog: str
    eventType: str | None = None
    formattedContent: str
    integrationType: IntegrationType
    metadata: dict[str, Any] | None = None


class SdkAgentRunRequestBody(_CamelModel):
    event: PartialSdkAgentRunEventPayload | None = None
    options: SdkAgentRunOptionsPayload | None = None
    prompt: str | None = None
    skills: list[SdkAgentSkillPayload] | None = None


class NormalizedRequest(_CamelModel):
    event: SdkAgentRunEventPayload
    options: Options
    prompt: str
    skills: list[SdkAgentSkillPayload]


class SdkAgentRunResponseBody(_CamelModel):
    contract: Contract | None = None
    details: list[str] | None = None
    error: str | None = None
    normalizedRequest: NormalizedRequest | None = None
    success: bool


__all__ = [
    "ApiToken",
    "ApiTokenCreateResponse",
    "Contract",
    "NormalizedRequest",
    "Options",
    "PartialSdkAgentRunEventPayload",
    "RemovedItem",
    "Result",
    "SdkAgentRunEventPayload",
    "SdkAgentRunOptionsPayload",
    "SdkAgentRunRequestBody",
    "SdkAgentRunResponseBody",
    "SdkAgentSkillPayload",
    "SdkDeployJob",
    "SdkDeployRequestBody",
    "SdkDeployResponseBody",
    "SdkDeployTrigger",
    "SerializedEvent",
    "TriggerPayload",
]



"""Agent configuration models."""

from __future__ import annotations

from typing import Literal

from ._base import _CamelModel
from .enums import RunHistoryActionType
from .integrations import ConfigInstance


class AgentPrompt(_CamelModel):
    text: str


class AgentNotificationSettings(_CamelModel):
    actionTypes: list[RunHistoryActionType]
    enabled: bool


class AgentOutput(_CamelModel):
    config: ConfigInstance
    id: str


class AgentTrigger(_CamelModel):
    config: ConfigInstance
    id: str


class AgentUpdate(_CamelModel):
    isActive: bool | None = None
    name: str | None = None
    notificationSettings: AgentNotificationSettings | None = None
    outputs: list[AgentOutput] | None = None
    prompt: AgentPrompt | None = None
    requireApproval: bool | None = None
    toolApprovals: list[str] | None = None
    triggers: list[AgentTrigger] | None = None


class Agent(_CamelModel):
    createdByUserId: str
    id: str
    isActive: bool
    name: str
    notificationSettings: AgentNotificationSettings | None = None
    outputs: list[AgentOutput]
    prompt: AgentPrompt
    requireApproval: bool
    source: Literal["WEB_UI", "SDK"] | None = None
    toolApprovals: list[str] | None = None
    triggers: list[AgentTrigger]
    updatedAt: str | None = None


__all__ = [
    "Agent",
    "AgentNotificationSettings",
    "AgentOutput",
    "AgentPrompt",
    "AgentTrigger",
    "AgentUpdate",
]



# ruff: noqa: F401

"""Terse Python SDK."""

from .runtime import (
    EventType,
    MissingApiKeyError,
    RegisteredJob,
    Terse,
    TerseAgent,
    TerseApiError,
    TerseRuntimeError,
    clear_job_registry,
    deserialize_input_event,
    execute_registered_job,
    get_job_registry,
)
from .types.agents import (
    Agent,
    AgentNotificationSettings,
    AgentOutput,
    AgentPrompt,
    AgentTrigger,
    AgentUpdate,
)
from .types.agents import __all__ as _agents_all
from .types.attio import (
    AttioAttribute,
    AttioObjectWithAttributes,
    AttioRecord,
    AttioRecordIdentifier,
    Repository,
)
from .types.attio import __all__ as _attio_all
from .types.chat_snippets import (
    ChatSnippet,
    ChatSnippetButton,
    ChatSnippetImage,
    ChatSnippetIntegrationPrompt,
    ChatSnippetMultipleChoice,
    ChatSnippetNavigate,
    MultipleChoiceOption,
)
from .types.chat_snippets import __all__ as _chat_snippets_all
from .types.config import TerseSettings
from .types.config import __all__ as _config_all
from .types.enums import (
    ApprovalActionType,
    ApprovalRequestStatus,
    ChangeEventType,
    ConfigType,
    EntityType,
    FigmaEventType,
    GitHubEventType,
    GmailEventType,
    IntegrationType,
    JiraEventType,
    LinearEventType,
    NotificationDestinationType,
    RunHistoryActionType,
    RunHistoryDecisionAction,
    RunHistoryStatus,
    SlackEventType,
    TicketSystemType,
    ToolCallExecutionStatus,
    WorkOSEventType,
)
from .types.enums import __all__ as _enums_all
from .types.events import (
    AnyInputEvent,
    AtlassianInputEvent,
    AttioInputEvent,
    CronJobInputEvent,
    DatadogInputEvent,
    FigmaInputEvent,
    GithubCommit,
    GithubEventMetadata,
    GithubFileDiff,
    GithubInputEvent,
    GithubPRData,
    GithubPRRef,
    GithubRepository,
    GithubUser,
    GmailInputEvent,
    InputEvent,
    KnownInputEvent,
    LaunchDarklyInputEvent,
    LinearInputEvent,
    NotionInputEvent,
    PosthogInputEvent,
    SerializedEventInputEvent,
    SlackInputEvent,
    SnowflakeInputEvent,
    TerseInputEvent,
    WorkOSEventInvitation,
    WorkOSEventMembership,
    WorkOSEventMetadata,
    WorkOSEventUser,
    WorkOSInputEvent,
)
from .types.events import __all__ as _events_all
from .types.integrations import (
    ConfigDetails,
    ConfigInstance,
    IntegrationDetails,
    NotificationSettings,
)
from .types.integrations import __all__ as _integrations_all
from .types.jobs import JobDefinition, SkillConfig, TriggerConfig
from .types.jobs import __all__ as _jobs_all
from .types.model_events import (
    Error,
    ModelEvent,
    ModelEventCancelled,
    ModelEventFilterResult,
    ModelEventNaturalStop,
    ModelEventRunError,
    ModelEventSnippet,
    ModelEventTextDelta,
    ModelEventThinking,
    ModelEventToolApprovalRequest,
    ModelEventToolApprovalResponse,
    ModelEventToolCall,
    ModelEventToolCallComplete,
    ModelEventToolCallGenerating,
    ModelEventUserMessage,
    RunError,
    SharedErrorContext,
    TextDelta,
    ToolCall,
)
from .types.model_events import __all__ as _model_events_all
from .types.run_history import (
    ApprovalAction,
    ApprovalRequest,
    ChangedItem,
    OutputItem,
    RunHistoryAction,
    RunHistoryDecision,
    RunHistoryRecord,
    RunHistoryTrigger,
)
from .types.run_history import __all__ as _run_history_all
from .types.sdk_types import (
    ApiToken,
    ApiTokenCreateResponse,
    Contract,
    NormalizedRequest,
    Options,
    PartialSdkAgentRunEventPayload,
    RemovedItem,
    Result,
    SdkAgentRunEventPayload,
    SdkAgentRunOptionsPayload,
    SdkAgentRunRequestBody,
    SdkAgentRunResponseBody,
    SdkAgentSkillPayload,
    SdkDeployJob,
    SdkDeployRequestBody,
    SdkDeployResponseBody,
    SdkDeployTrigger,
    SerializedEvent,
    TriggerPayload,
)
from .types.sdk_types import __all__ as _sdk_types_all
from .types.stream_events import (
    SdkAgentStreamEvent,
    SdkAgentStreamEventAction,
    SdkAgentStreamEventDone,
    SdkAgentStreamEventError,
    SdkAgentStreamEventFinalOutput,
    SdkAgentStreamEventText,
    SdkAgentStreamEventToolCallCompleted,
    SdkAgentStreamEventToolCallParams,
    SdkAgentStreamEventToolCallStarted,
)
from .types.stream_events import __all__ as _stream_events_all
from .types.tools import (
    AttioListObjectsToolOutput,
    AttioQueryRecordsToolOutput,
    AttioUpsertRecordToolOutput,
    SnowflakeExecuteQueryToolOutput,
    SnowflakeExplainQueryToolOutput,
    ToolOutputBase,
)
from .types.tools import __all__ as _tools_all
from .types.users import Role, User, UserNoOrganization
from .types.users import __all__ as _users_all

__version__ = "0.1.3"

__all__ = [
    "__version__",
    *_agents_all,
    *_attio_all,
    *_chat_snippets_all,
    *_config_all,
    *_enums_all,
    *_events_all,
    *_integrations_all,
    *_jobs_all,
    *_model_events_all,
    *_run_history_all,
    *_sdk_types_all,
    *_stream_events_all,
    *_tools_all,
    *_users_all,
    "EventType",
    "MissingApiKeyError",
    "RegisteredJob",
    "Terse",
    "TerseAgent",
    "TerseApiError",
    "TerseRuntimeError",
    "clear_job_registry",
    "deserialize_input_event",
    "execute_registered_job",
    "get_job_registry",
]



"""Typed settings loaded from `TERSE_*` environment variables."""

from pydantic_settings import BaseSettings, SettingsConfigDict


class TerseSettings(BaseSettings):
    api_key: str = ""
    frontend_url: str = "https://app.useterse.ai"
    backend_url: str = "https://cursor-for-tickets.onrender.com"

    model_config = SettingsConfigDict(env_prefix="TERSE_", extra="ignore")


__all__ = ["TerseSettings"]



"""Enum-like literal aliases shared across SDK DTOs."""

from __future__ import annotations

from typing import Literal, TypeAlias

ApprovalActionType: TypeAlias = Literal["open_run_history", "approve_action", "reject_action"]
ApprovalRequestStatus: TypeAlias = Literal["pending", "in_progress", "completed"]
ChangeEventType: TypeAlias = Literal["CREATED", "UPDATED", "ACTION_EXECUTED"]
ConfigType: TypeAlias = Literal[
    "gmail",
    "gmail_output",
    "gmail_draft_output",
    "figma",
    "slack",
    "slack_output",
    "notion",
    "linear_input",
    "linear_output",
    "github",
    "jira",
    "confluence",
    "POSTHOG",
    "DATADOG",
    "time_trigger",
    "launchdarkly",
    "terse",
    "workos_input",
    "workos_output",
    "attio_output",
    "snowflake_output",
]
EntityType: TypeAlias = Literal["ticket", "comment", "user", "action_event", "run_history_action"]
FigmaEventType: TypeAlias = Literal["file_comment"]
GitHubEventType: TypeAlias = Literal[
    "push",
    "pull_request.opened",
    "pull_request.merged",
    "pull_request.closed",
    "pull_request.synchronize",
]
GmailEventType: TypeAlias = Literal["email.received"]
IntegrationType: TypeAlias = Literal[
    "github",
    "gmail",
    "linear",
    "atlassian",
    "slack",
    "notion",
    "figma",
    "terse",
    "posthog",
    "datadog",
    "cron_job",
    "launchdarkly",
    "workos",
    "attio",
    "snowflake",
]
JiraEventType: TypeAlias = Literal["issue.created", "issue.updated"]
LinearEventType: TypeAlias = Literal["issue.created", "issue.updated", "comment.created"]
NotificationDestinationType: TypeAlias = Literal["email", "slack"]
RunHistoryActionType: TypeAlias = Literal["create", "update", "delete", "read", "approve", "error"]
RunHistoryDecisionAction: TypeAlias = Literal["processed", "skipped"]
RunHistoryStatus: TypeAlias = Literal["success", "failed", "cancelled", "skipped", "in_progress", "awaiting_approval"]
SlackEventType: TypeAlias = Literal["message", "app_mention", "reaction_added"]
TicketSystemType: TypeAlias = Literal["jira", "linear"]
ToolCallExecutionStatus: TypeAlias = Literal["completed", "incomplete", "failed", "unknown"]
WorkOSEventType: TypeAlias = Literal[
    "user.created",
    "user.updated",
    "user.deleted",
    "organization.created",
    "organization_membership.created",
    "organization_membership.updated",
    "organization_membership.deleted",
    "invitation.created",
    "invitation.accepted",
    "invitation.resent",
    "invitation.revoked",
]

__all__ = [
    "ApprovalActionType",
    "ApprovalRequestStatus",
    "ChangeEventType",
    "ConfigType",
    "EntityType",
    "FigmaEventType",
    "GitHubEventType",
    "GmailEventType",
    "IntegrationType",
    "JiraEventType",
    "LinearEventType",
    "NotificationDestinationType",
    "RunHistoryActionType",
    "RunHistoryDecisionAction",
    "RunHistoryStatus",
    "SlackEventType",
    "TicketSystemType",
    "ToolCallExecutionStatus",
    "WorkOSEventType",
]



"""User models for the Python SDK."""

from __future__ import annotations

from typing import Literal, TypeAlias

from ._base import _CamelModel

Role: TypeAlias = Literal["admin", "user"]


class User(_CamelModel):
    displayName: str
    displayPhotoUrl: str
    email: str
    firstName: str | None
    id: str
    lastName: str | None
    organizationId: str
    organizationName: str
    roles: list[Role]
    workosId: str


class UserNoOrganization(_CamelModel):
    displayName: str
    displayPhotoUrl: str
    email: str
    firstName: str | None
    id: str
    lastName: str | None
    workosId: str


__all__ = ["Role", "User", "UserNoOrganization"]



"""Chat snippet models."""

from __future__ import annotations

from typing import Annotated, Literal

from pydantic import Field, RootModel

from ._base import _CamelModel


class MultipleChoiceOption(_CamelModel):
    label: str
    value: str


class ChatSnippetButton(_CamelModel):
    id: str | None = None
    label: str
    selectedValue: str | None = None
    step_id: str | None = None
    type: Literal["button"]
    url: str


class ChatSnippetIntegrationPrompt(_CamelModel):
    id: str | None = None
    integration: str
    message: str
    selectedValue: str | None = None
    stateToken: str | None = None
    step_id: str | None = None
    type: Literal["integration_prompt"]


class ChatSnippetNavigate(_CamelModel):
    id: str | None = None
    path: str
    selectedValue: str | None = None
    step_id: str | None = None
    type: Literal["navigate"]


class ChatSnippetMultipleChoice(_CamelModel):
    allowMultiple: bool | None = None
    id: str | None = None
    options: list[MultipleChoiceOption]
    question: str
    questionId: str
    selectedValue: str | None = None
    step_id: str | None = None
    type: Literal["multiple_choice"]


class ChatSnippetImage(_CamelModel):
    id: str | None = None
    selectedValue: str | None = None
    step_id: str | None = None
    type: Literal["image"]
    url: str


class ChatSnippet(
    RootModel[
        ChatSnippetButton
        | ChatSnippetIntegrationPrompt
        | ChatSnippetNavigate
        | ChatSnippetMultipleChoice
        | ChatSnippetImage
    ]
):
    root: Annotated[
        ChatSnippetButton
        | ChatSnippetIntegrationPrompt
        | ChatSnippetNavigate
        | ChatSnippetMultipleChoice
        | ChatSnippetImage,
        Field(discriminator="type"),
    ]


__all__ = [
    "ChatSnippet",
    "ChatSnippetButton",
    "ChatSnippetImage",
    "ChatSnippetIntegrationPrompt",
    "ChatSnippetMultipleChoice",
    "ChatSnippetNavigate",
    "MultipleChoiceOption",
]



"""Streaming event models for SDK agent runs."""

from __future__ import annotations

from typing import Annotated, Literal

from pydantic import Field, RootModel

from ._base import _CamelModel
from .run_history import RunHistoryAction


class SdkAgentStreamEventText(_CamelModel):
    text: str
    type: Literal["text"]


class SdkAgentStreamEventFinalOutput(_CamelModel):
    finalOutput: str
    type: Literal["final_output"]


class SdkAgentStreamEventToolCallParams(_CamelModel):
    toolCallParams: str
    type: Literal["tool_call_params"]


class SdkAgentStreamEventToolCallStarted(_CamelModel):
    toolCallStarted: str
    type: Literal["tool_call_started"]


class SdkAgentStreamEventToolCallCompleted(_CamelModel):
    toolCallCompleted: str
    type: Literal["tool_call_completed"]


class SdkAgentStreamEventAction(_CamelModel):
    action: RunHistoryAction
    type: Literal["action"]


class SdkAgentStreamEventError(_CamelModel):
    message: str
    type: Literal["error"]


class SdkAgentStreamEventDone(_CamelModel):
    type: Literal["done"]


class SdkAgentStreamEvent(
    RootModel[
        SdkAgentStreamEventText
        | SdkAgentStreamEventFinalOutput
        | SdkAgentStreamEventToolCallParams
        | SdkAgentStreamEventToolCallStarted
        | SdkAgentStreamEventToolCallCompleted
        | SdkAgentStreamEventAction
        | SdkAgentStreamEventError
        | SdkAgentStreamEventDone
    ]
):
    root: Annotated[
        SdkAgentStreamEventText
        | SdkAgentStreamEventFinalOutput
        | SdkAgentStreamEventToolCallParams
        | SdkAgentStreamEventToolCallStarted
        | SdkAgentStreamEventToolCallCompleted
        | SdkAgentStreamEventAction
        | SdkAgentStreamEventError
        | SdkAgentStreamEventDone,
        Field(discriminator="type"),
    ]


__all__ = [
    "SdkAgentStreamEvent",
    "SdkAgentStreamEventAction",
    "SdkAgentStreamEventDone",
    "SdkAgentStreamEventError",
    "SdkAgentStreamEventFinalOutput",
    "SdkAgentStreamEventText",
    "SdkAgentStreamEventToolCallCompleted",
    "SdkAgentStreamEventToolCallParams",
    "SdkAgentStreamEventToolCallStarted",
]



"""Job definition models for the Python SDK."""

from __future__ import annotations

from typing import Any

from pydantic import Field

from ._base import TerseModel


class TriggerConfig(TerseModel):
    integration_id: str
    integration_type: str
    event_type: str | None = None
    config_type: str | None = None
    config: dict[str, Any] = Field(default_factory=dict)


class SkillConfig(TerseModel):
    integration_id: str
    integration_type: str
    config_type: str
    config: dict[str, Any] = Field(default_factory=dict)


class JobDefinition(TerseModel):
    name: str
    triggers: list[TriggerConfig] = Field(default_factory=list)
    skills: list[SkillConfig] = Field(default_factory=list)


__all__ = ["JobDefinition", "SkillConfig", "TriggerConfig"]



"""Shared model helpers for the Python SDK."""

from pydantic import BaseModel, ConfigDict


def to_camel(value: str) -> str:
    head, *tail = value.split("_")
    return head + "".join(part.capitalize() for part in tail)


class TerseModel(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel, extra="forbid", populate_by_name=True)


class _CamelModel(BaseModel):
    model_config = ConfigDict(extra="forbid")



"""Integration and config models."""

from __future__ import annotations

from ._base import _CamelModel
from .enums import ConfigType, IntegrationType, RunHistoryActionType


class IntegrationDetails(_CamelModel):
    description: str
    isInput: bool | None = None
    isOutput: bool | None = None
    name: str
    type: IntegrationType


class ConfigDetails(_CamelModel):
    configType: ConfigType
    description: str
    integrationType: IntegrationType
    isInput: bool
    isOutput: bool
    name: str


class ConfigInstance(_CamelModel):
    configType: ConfigType
    integrationId: str
    integrationType: IntegrationType


class NotificationSettings(_CamelModel):
    agentDefaultNotifications: list[RunHistoryActionType]
    id: str
    weeklyAgentImprovements: bool


__all__ = ["ConfigDetails", "ConfigInstance", "IntegrationDetails", "NotificationSettings"]



"""Tool output models."""

from __future__ import annotations

from typing import Any

from ._base import _CamelModel
from .attio import AttioObjectWithAttributes, AttioRecord
from .run_history import RunHistoryAction


class ToolOutputBase(_CamelModel):
    actions: list[RunHistoryAction] | None = None
    success: bool


class AttioListObjectsToolOutput(ToolOutputBase):
    count: float
    objects: list[AttioObjectWithAttributes]


class AttioQueryRecordsToolOutput(ToolOutputBase):
    count: float
    records: list[AttioRecord]


class AttioUpsertRecordToolOutput(ToolOutputBase):
    record: AttioRecord | None = None


class SnowflakeExecuteQueryToolOutput(ToolOutputBase):
    columns: list[str]
    rowCount: float
    rows: list[dict[str, Any]]


class SnowflakeExplainQueryToolOutput(ToolOutputBase):
    columns: list[str]
    explainPlan: list[dict[str, Any]]
    rowCount: float


__all__ = [
    "AttioListObjectsToolOutput",
    "AttioQueryRecordsToolOutput",
    "AttioUpsertRecordToolOutput",
    "SnowflakeExecuteQueryToolOutput",
    "SnowflakeExplainQueryToolOutput",
    "ToolOutputBase",
]



"""Runtime helpers for registering and executing Terse Python jobs."""

from __future__ import annotations

import asyncio
import inspect
import json
import logging
import os
import sys
from collections.abc import Awaitable, Callable, Coroutine, Generator, Mapping, Sequence
from dataclasses import dataclass, field
from enum import StrEnum
from typing import Any, TypeVar, cast

import httpx
from httpx_sse import connect_sse
from pydantic import ValidationError

from .types.config import TerseSettings
from .types.events import (
    AnyInputEvent,
    AtlassianInputEvent,
    AttioInputEvent,
    CronJobInputEvent,
    DatadogInputEvent,
    FigmaInputEvent,
    GithubInputEvent,
    GmailInputEvent,
    InputEvent,
    LaunchDarklyInputEvent,
    LinearInputEvent,
    NotionInputEvent,
    PosthogInputEvent,
    SerializedEventInputEvent,
    SlackInputEvent,
    SnowflakeInputEvent,
    TerseInputEvent,
    WorkOSInputEvent,
)
from .types.jobs import SkillConfig, TriggerConfig
from .types.sdk_types import (
    SdkAgentRunRequestBody,
    SdkAgentRunResponseBody,
    SdkAgentSkillPayload,
    SerializedEvent,
)
from .types.stream_events import (
    SdkAgentStreamEvent,
    SdkAgentStreamEventAction,
    SdkAgentStreamEventDone,
    SdkAgentStreamEventError,
    SdkAgentStreamEventFinalOutput,
    SdkAgentStreamEventText,
    SdkAgentStreamEventToolCallCompleted,
    SdkAgentStreamEventToolCallParams,
    SdkAgentStreamEventToolCallStarted,
)

JobEvent = AnyInputEvent
JobHandler = Callable[[JobEvent, "TerseAgent"], None | Awaitable[None]]
JobFilter = Callable[[JobEvent], bool | Awaitable[bool]]
AgentStreamEvent = (
    SdkAgentStreamEventText
    | SdkAgentStreamEventFinalOutput
    | SdkAgentStreamEventToolCallParams
    | SdkAgentStreamEventToolCallStarted
    | SdkAgentStreamEventToolCallCompleted
    | SdkAgentStreamEventAction
)

HandlerT = TypeVar("HandlerT", bound=Callable[..., object])
ResultT = TypeVar("ResultT")

_JOB_REGISTRY: dict[str, RegisteredJob] = {}
LOGGER = logging.getLogger("terse.sdk.runtime")
_EVENT_MODEL_BY_INTEGRATION = {
    "github": GithubInputEvent,
    "slack": SlackInputEvent,
    "linear": LinearInputEvent,
    "atlassian": AtlassianInputEvent,
    "gmail": GmailInputEvent,
    "notion": NotionInputEvent,
    "figma": FigmaInputEvent,
    "posthog": PosthogInputEvent,
    "datadog": DatadogInputEvent,
    "terse": TerseInputEvent,
    "cron_job": CronJobInputEvent,
    "launchdarkly": LaunchDarklyInputEvent,
    "workos": WorkOSInputEvent,
    "attio": AttioInputEvent,
    "snowflake": SnowflakeInputEvent,
}


class TerseRuntimeError(RuntimeError):
    """Base runtime error for the Python SDK."""


class MissingApiKeyError(TerseRuntimeError):
    """Raised when a command requires ``TERSE_API_KEY`` and it is missing."""


class TerseApiError(TerseRuntimeError):
    """Raised when a backend request fails."""


class EventType(StrEnum):
    """Stream event type constants for agent runs."""

    TEXT = "text"
    FINAL_OUTPUT = "final_output"
    TOOL_CALL_PARAMS = "tool_call_params"
    TOOL_CALL_STARTED = "tool_call_started"
    TOOL_CALL_COMPLETED = "tool_call_completed"
    ACTION = "action"


@dataclass(frozen=True)
class RegisteredJob:
    """A runtime job registration captured from ``@app.job``."""

    name: str
    handler: Callable[..., object] = field(repr=False, compare=False)
    triggers: list[TriggerConfig] = field(default_factory=list)
    skills: list[SkillConfig] = field(default_factory=list)
    filter: JobFilter | None = field(default=None, repr=False, compare=False)
    webhook_url: str | None = None


class Terse:
    """Python entry point for registering jobs."""

    def job(
        self,
        name: str,
        triggers: Sequence[TriggerConfig] | None = None,
        skills: Sequence[SkillConfig] | None = None,
        filter: JobFilter | None = None,
        webhook_url: str | None = None,
    ) -> Callable[[HandlerT], HandlerT]:
        """Register a job and return the original handler."""

        def decorator(handler: HandlerT) -> HandlerT:
            _JOB_REGISTRY[name] = RegisteredJob(
                name=name,
                handler=handler,
                triggers=list(triggers or []),
                skills=list(skills or []),
                filter=filter,
                webhook_url=webhook_url,
            )
            return handler

        return decorator


class TerseAgent:
    """Small Python runtime client for agent runs and deterministic tool calls."""

    def __init__(
        self,
        skills: Sequence[SkillConfig] | None = None,
        backend_url: str | None = None,
        session_id: str | None = None,
    ) -> None:
        settings = TerseSettings()
        self.skills = list(skills or [])
        self.backend_url = (backend_url or settings.backend_url).rstrip("/")
        self.session_id = session_id
        self._tools: object | None = None

    @property
    def tools(self) -> object:
        """Return generated deterministic tool wrappers for this agent."""

        if self._tools is None:
            self.ensure_generated_tools()
        if self._tools is None:
            raise AttributeError(
                "No generated tools are attached. Run `terse generate` and import `terse_generated` in your project."
            )
        return self._tools

    @tools.setter
    def tools(self, value: object) -> None:
        self._tools = value

    def attach_tools(self, tools: object) -> object:
        """Attach generated tool wrappers to the agent and return them."""

        self._tools = tools
        return tools

    def ensure_generated_tools(self) -> object | None:
        """Attach generated tool wrappers if a project-local factory is loaded."""

        if self._tools is not None:
            return self._tools

        factory = _resolve_generated_tools_factory()
        if factory is None:
            return None

        self._tools = factory(self)
        return self._tools

    def run(self, prompt: str, event: InputEvent | None = None) -> Generator[AgentStreamEvent, None, None]:
        """Stream parsed agent-run events from the backend."""

        api_key = _require_api_key()
        request_body = SdkAgentRunRequestBody.model_validate(
            {
                "prompt": prompt,
                "event": _serialize_run_event(event or _manual_event()),
                "skills": [_serialize_skill_config(skill).model_dump(exclude_none=True) for skill in self.skills],
            }
        )
        request_payload = request_body.model_dump(exclude_none=True)
        headers = _build_auth_headers(api_key, accept="text/event-stream", session_id=self.session_id)
        _debug_log_request("POST", f"{self.backend_url}/sdk/agent-run", headers, request_payload)
        failed_tool_calls: list[str] = []

        try:
            with (
                httpx.Client(timeout=None) as client,
                connect_sse(
                    client,
                    "POST",
                    f"{self.backend_url}/sdk/agent-run",
                    headers=headers,
                    json=request_payload,
                ) as event_source,
            ):
                _assert_sse_response(event_source.response, "/sdk/agent-run")

                for sse in event_source.iter_sse():
                    if not sse.data:
                        continue

                    stream_event = SdkAgentStreamEvent.model_validate_json(sse.data).root
                    if isinstance(stream_event, SdkAgentStreamEventDone):
                        if failed_tool_calls:
                            raise TerseApiError(f"Run completed with failed tool calls: {'; '.join(failed_tool_calls)}")
                        return
                    if isinstance(stream_event, SdkAgentStreamEventError):
                        raise TerseApiError(stream_event.message)
                    if isinstance(stream_event, SdkAgentStreamEventToolCallCompleted):
                        parsed = _parse_tool_call_completed(stream_event.toolCallCompleted)
                        if parsed.get("status") and parsed["status"] != "completed":
                            failed_tool_calls.append(f"{parsed.get('tool', 'unknown_tool')}: {parsed['status']}")
                    yield cast(AgentStreamEvent, stream_event)
        except httpx.RequestError as exc:
            raise TerseApiError(f"Could not connect to {self.backend_url} — is the backend running?\n  {exc}") from exc
        except ValidationError as exc:
            raise TerseApiError(f"Received invalid agent stream payload.\n  {exc}") from exc

    def run_and_wait(self, prompt: str, event: InputEvent | None = None) -> str | None:
        """Run the agent to completion and return the final output, if any."""

        final_output: str | None = None
        for chunk in self.run(prompt, event):
            if isinstance(chunk, SdkAgentStreamEventFinalOutput):
                final_output = chunk.finalOutput
        return final_output

    def execute_tool(self, tool_name: str, params: Mapping[str, object] | None = None) -> object:
        """Execute a deterministic tool via the backend."""

        api_key = _require_api_key()
        headers = _build_auth_headers(api_key, session_id=self.session_id)
        request_payload = {"toolName": tool_name, "params": dict(params or {})}
        _debug_log_request("POST", f"{self.backend_url}/sdk/tool-execute", headers, request_payload)

        try:
            with httpx.Client(timeout=20.0) as client:
                response = client.post(
                    f"{self.backend_url}/sdk/tool-execute",
                    headers=headers,
                    json=request_payload,
                )
        except httpx.RequestError as exc:
            raise TerseApiError(f"Could not connect to {self.backend_url} — is the backend running?\n  {exc}") from exc

        payload = _read_json_response(response, "/sdk/tool-execute")
        _debug_log_response_payload("/sdk/tool-execute", payload)
        payload_dict = _as_object_dict(payload)
        if response.is_error:
            detail = payload_dict.get("error") if payload_dict is not None else None
            raise TerseApiError(
                f"{response.status_code} {response.reason_phrase} — /sdk/tool-execute"
                + (f"\n  {detail}" if detail else "")
            )

        if payload_dict is None or payload_dict.get("success") is not True:
            detail = payload_dict.get("error") if payload_dict is not None else None
            raise TerseApiError(str(detail or "Tool execution failed"))

        return payload_dict.get("result")


def clear_job_registry() -> None:
    """Clear the global job registry."""

    _JOB_REGISTRY.clear()


def get_job_registry() -> dict[str, RegisteredJob]:
    """Return a snapshot of the current job registry."""

    return dict(_JOB_REGISTRY)


def deserialize_input_event(
    value: SerializedEvent | Mapping[str, object] | str,
) -> AnyInputEvent:
    """Convert a serialized backend event into the best matching SDK event model."""

    if isinstance(value, SerializedEvent):
        payload = value.model_dump(exclude_none=True)
    elif isinstance(value, str):
        parsed = json.loads(value)
        if not isinstance(parsed, dict):
            raise TerseRuntimeError("Serialized event JSON must be an object.")
        payload = dict(parsed)
    else:
        payload = dict(value)

    event_class = _EVENT_MODEL_BY_INTEGRATION.get(_read_integration_type(payload))
    if event_class is None:
        return SerializedEventInputEvent.model_validate(payload)
    try:
        return event_class.model_validate(payload)
    except ValidationError:
        return SerializedEventInputEvent.model_validate(payload)


def execute_registered_job(
    job: RegisteredJob,
    event: AnyInputEvent,
    agent: TerseAgent | None = None,
) -> bool:
    """Execute a registered job and return ``True`` when it was skipped by the filter."""

    runtime_agent = agent or TerseAgent(job.skills)
    runtime_agent.ensure_generated_tools()

    if job.filter is not None:
        should_run = _run_callable(job.filter, event)
        if not should_run:
            return True

    _run_callable(job.handler, event, runtime_agent)
    return False


def _serialize_skill_config(skill: SkillConfig) -> SdkAgentSkillPayload:
    config = dict(skill.config)
    config["integrationId"] = skill.integration_id
    config["integrationType"] = skill.integration_type
    config["configType"] = skill.config_type
    return SdkAgentSkillPayload.model_validate({"configType": skill.config_type, "config": config})


def _resolve_generated_tools_factory() -> Callable[[TerseAgent], object] | None:
    module = sys.modules.get("terse_generated")
    factory = getattr(module, "create_tools", None) if module is not None else None
    if callable(factory):
        return cast(Callable[[TerseAgent], object], factory)

    for loaded_module in list(sys.modules.values()):
        module_file = getattr(loaded_module, "__file__", None)
        if not isinstance(module_file, str) or not module_file.endswith("/terse_generated.py"):
            continue
        factory = getattr(loaded_module, "create_tools", None)
        if callable(factory):
            return cast(Callable[[TerseAgent], object], factory)

    return None


def _serialize_run_event(event: InputEvent) -> dict[str, str]:
    return {
        "integrationType": event.integration_type,
        "formattedContent": event.formatted_content,
        "debugLog": event.debug_log,
    }


def _manual_event() -> TerseInputEvent:
    return TerseInputEvent(
        event_type="manual",
        formatted_content="Manual trigger from terse run",
        debug_log="[MockInputEvent] Manual trigger via SDK",
    )


def _build_auth_headers(
    api_key: str,
    accept: str = "application/json",
    session_id: str | None = None,
) -> dict[str, str]:
    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json",
        "Accept": accept,
    }

    if session_id:
        headers["X-Terse-Session-Id"] = session_id
    if run_id := os.environ.get("TERSE_RUN_ID"):
        headers["X-Terse-Run-Id"] = run_id

    return headers


def _assert_sse_response(response: httpx.Response, path: str) -> None:
    _debug_log_response_metadata(response, path)
    if response.is_error:
        detail = _read_response_detail(response)
        if detail:
            LOGGER.debug("Response detail from %s:\n%s", path, detail)
        raise TerseApiError(
            f"{response.status_code} {response.reason_phrase} — {path}" + (f"\n  {detail}" if detail else "")
        )

    content_type = response.headers.get("Content-Type", "")
    if "text/event-stream" in content_type:
        return

    payload = _read_json_response(response, path)
    _debug_log_response_payload(path, payload)
    if isinstance(payload, dict):
        response_body = SdkAgentRunResponseBody.model_validate(payload)
        if response_body.error:
            details = f" ({'; '.join(response_body.details)})" if response_body.details else ""
            raise TerseApiError(f"{response_body.error}{details}")

    raise TerseApiError(f"Expected text/event-stream from {path} but got {content_type or 'unknown content-type'}.")


def _read_json_response(response: httpx.Response, path: str) -> object:
    _buffer_response_content(response)
    try:
        return response.json()
    except ValueError as exc:
        raise TerseApiError(f"Received invalid JSON from {path}.") from exc


def _read_response_detail(response: httpx.Response) -> str:
    _buffer_response_content(response)
    try:
        payload = response.json()
    except ValueError:
        return response.text.strip()

    if isinstance(payload, dict):
        detail = payload.get("error")
        if detail is not None:
            return str(detail)
    return response.text.strip()


def _buffer_response_content(response: httpx.Response) -> None:
    response.read()


def _debug_log_request(method: str, url: str, headers: Mapping[str, str], payload: object | None) -> None:
    if not LOGGER.isEnabledFor(logging.DEBUG):
        return

    LOGGER.debug("HTTP %s %s", method.upper(), url)
    LOGGER.debug("Request headers:\n%s", _format_debug_value(_redact_headers(headers)))
    if payload is not None:
        LOGGER.debug("Request payload:\n%s", _format_debug_value(payload))


def _debug_log_response_metadata(response: httpx.Response, path: str) -> None:
    if not LOGGER.isEnabledFor(logging.DEBUG):
        return

    LOGGER.debug("Response %s %s for %s", response.status_code, response.reason_phrase, path)
    LOGGER.debug("Response headers:\n%s", _format_debug_value(dict(response.headers)))


def _debug_log_response_payload(path: str, payload: object) -> None:
    if not LOGGER.isEnabledFor(logging.DEBUG):
        return

    LOGGER.debug("Response payload from %s:\n%s", path, _format_debug_value(payload))


def _redact_headers(headers: Mapping[str, str]) -> dict[str, str]:
    redacted: dict[str, str] = {}
    for key, value in headers.items():
        if key.lower() == "authorization":
            redacted[key] = "Bearer ***"
            continue
        redacted[key] = value
    return redacted


def _format_debug_value(value: object) -> str:
    try:
        return json.dumps(value, indent=2, sort_keys=True, default=str)
    except TypeError:
        return str(value)


def _as_object_dict(value: object) -> dict[str, object] | None:
    if isinstance(value, dict):
        return cast(dict[str, object], value)
    return None


def _read_integration_type(payload: Mapping[str, object]) -> str:
    raw_value = payload.get("integrationType", payload.get("integration_type"))
    return str(raw_value or "")


def _parse_tool_call_completed(raw: str) -> dict[str, str]:
    try:
        payload = json.loads(raw)
    except json.JSONDecodeError:
        return {}

    if isinstance(payload, dict):
        return {str(key): str(value) for key, value in payload.items() if value is not None}
    return {}


def _require_api_key() -> str:
    api_key = TerseSettings().api_key.strip()
    if not api_key:
        raise MissingApiKeyError(
            "TERSE_API_KEY environment variable is not set. "
            "Add it to your .env file or export it before running this command."
        )
    return api_key


def _run_callable(func: Callable[..., ResultT | Awaitable[ResultT]], *args: object) -> ResultT:
    result = func(*args)
    if inspect.isawaitable(result):
        return _await_result(cast(Awaitable[ResultT], result))
    return cast(ResultT, result)


def _await_result(awaitable: Awaitable[ResultT]) -> ResultT:
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.run(cast(Coroutine[Any, Any, ResultT], awaitable))
    raise TerseRuntimeError("Cannot synchronously await a coroutine while another event loop is running.")



"""Model event DTOs."""

from __future__ import annotations

from typing import Annotated, Any, Literal

from pydantic import Field, RootModel

from ._base import _CamelModel
from .chat_snippets import ChatSnippet
from .enums import ToolCallExecutionStatus
from .run_history import ChangedItem


class Error(_CamelModel):
    message: str
    name: str
    stack: str | None = None


class RunError(_CamelModel):
    code: str | None = None
    error: str
    timestamp: float


class ToolCall(_CamelModel):
    integration: str
    parameters: str
    step_id: str
    summary: str
    timestamp: float


class TextDelta(_CamelModel):
    delta: str
    step_id: str
    timestamp: float


class SharedErrorContext(_CamelModel):
    error: Error | Any


class ModelEventToolApprovalResponse(_CamelModel):
    approved: bool
    step_id: str
    timestamp: float
    type: Literal["ToolApprovalResponse"]


class ModelEventToolApprovalRequest(_CamelModel):
    arguments: str
    name: str
    step_id: str
    timestamp: float
    type: Literal["ToolApprovalRequest"]


class ModelEventToolCallGenerating(_CamelModel):
    step_id: str
    timestamp: float
    tool_name: str
    type: Literal["ToolCallGenerating"]


class ModelEventToolCall(_CamelModel):
    integration: str
    parameters: str
    step_id: str
    summary: str
    timestamp: float
    type: Literal["ToolCall"]


class ModelEventToolCallComplete(_CamelModel):
    changed_items: list[ChangedItem]
    errorContext: SharedErrorContext | None = None
    integration: str
    result: str | None = None
    status: ToolCallExecutionStatus
    step_id: str
    timestamp: float
    tool_name: str
    type: Literal["ToolCallComplete"]
    url: str | None = None


class ModelEventTextDelta(_CamelModel):
    delta: str
    step_id: str
    timestamp: float
    type: Literal["TextDelta"]


class ModelEventRunError(_CamelModel):
    code: str | None = None
    error: str
    timestamp: float
    type: Literal["RunError"]


class ModelEventCancelled(_CamelModel):
    reason: str | None = None
    timestamp: float
    type: Literal["Cancelled"]


class ModelEventNaturalStop(_CamelModel):
    step_id: str
    timestamp: float
    type: Literal["NaturalStop"]


class ModelEventFilterResult(_CamelModel):
    confidence: float
    isRelevant: bool
    reason: str
    step_id: str
    timestamp: float
    type: Literal["FilterResult"]


class ModelEventUserMessage(_CamelModel):
    client_turn_id: str
    message: str
    step_id: str
    timestamp: float
    type: Literal["UserMessage"]


class ModelEventThinking(_CamelModel):
    step_id: str
    timestamp: float
    type: Literal["Thinking"]


class ModelEventSnippet(_CamelModel):
    snippet: ChatSnippet
    timestamp: float
    type: Literal["Snippet"]


class ModelEvent(
    RootModel[
        ModelEventToolApprovalResponse
        | ModelEventToolApprovalRequest
        | ModelEventToolCallGenerating
        | ModelEventToolCall
        | ModelEventToolCallComplete
        | ModelEventTextDelta
        | ModelEventRunError
        | ModelEventCancelled
        | ModelEventNaturalStop
        | ModelEventFilterResult
        | ModelEventUserMessage
        | ModelEventThinking
        | ModelEventSnippet
    ]
):
    root: Annotated[
        ModelEventToolApprovalResponse
        | ModelEventToolApprovalRequest
        | ModelEventToolCallGenerating
        | ModelEventToolCall
        | ModelEventToolCallComplete
        | ModelEventTextDelta
        | ModelEventRunError
        | ModelEventCancelled
        | ModelEventNaturalStop
        | ModelEventFilterResult
        | ModelEventUserMessage
        | ModelEventThinking
        | ModelEventSnippet,
        Field(discriminator="type"),
    ]


__all__ = [
    "Error",
    "ModelEvent",
    "ModelEventCancelled",
    "ModelEventFilterResult",
    "ModelEventNaturalStop",
    "ModelEventRunError",
    "ModelEventSnippet",
    "ModelEventTextDelta",
    "ModelEventThinking",
    "ModelEventToolApprovalRequest",
    "ModelEventToolApprovalResponse",
    "ModelEventToolCall",
    "ModelEventToolCallComplete",
    "ModelEventToolCallGenerating",
    "ModelEventUserMessage",
    "RunError",
    "SharedErrorContext",
    "TextDelta",
    "ToolCall",
]
